local rdebug = require 'remotedebug.visitor'
local json = require 'common.json'
local variables = require 'backend.worker.variables'
local source = require 'backend.worker.source'
local breakpoint = require 'backend.worker.breakpoint'
local evaluate = require 'backend.worker.evaluate'
local traceback = require 'backend.worker.traceback'
local stdout = require 'backend.worker.stdout'
local luaver = require 'backend.worker.luaver'
local ev = require 'backend.event'
local hookmgr = require 'remotedebug.hookmgr'
local stdio = require 'remotedebug.stdio'
local thread = require 'remotedebug.thread'
local fs = require 'backend.worker.filesystem'
local err = thread.channel 'errlog'

local initialized = false
local info = {}
local state = 'running'
local stopReason = 'step'
local currentException = {
    message = '',
    trace = '',
}
local outputCapture = {}
local noDebug = false
local openUpdate = true
local coroutineTree = {}
local stackFrame = {}
local skipFrame = 0
local baseL

local CMD = {}

local WorkerIdent = tostring(thread.id)
local WorkerChannel = ('DbgWorker(%s)'):format(WorkerIdent)

thread.newchannel(WorkerChannel)
local masterThread = thread.channel 'DbgMaster'
local workerThread = thread.channel (WorkerChannel)

local function workerThreadUpdate(timeout)
    while true do
        local ok, msg = workerThread:pop(timeout)
        if not ok then
            break
        end
        local ok, e = xpcall(function()
            local pkg = json.decode(msg)
            local f = CMD[pkg.cmd]
            if f then
                f(pkg)
            end
        end, debug.traceback)
        if not ok then
            err:push(e)
        end
    end
end

local function sendToMaster(cmd)
    return function (pkg)
        masterThread:push(WorkerIdent, cmd, json.encode(pkg))
    end
end

ev.on('breakpoint', function(reason, bp)
    sendToMaster 'eventBreakpoint' {
        reason = reason,
        breakpoint = bp,
    }
end)

ev.on('output', function(body)
    sendToMaster 'eventOutput' (body)
end)

ev.on('loadedSource', function(reason, s)
    sendToMaster 'eventLoadedSource' {
        reason = reason,
        source = source.output(s)
    }
end)

--function print(...)
--    local n = select('#', ...)
--    local t = {}
--    for i = 1, n do
--        t[i] = tostring(select(i, ...))
--    end
--    ev.emit('output', {
--        category = 'stderr',
--        output = table.concat(t, '\t')..'\n',
--    })
--end

--local log = require 'common.log'
--print = log.info

local function cleanFrame()
    variables.clean()
    stackFrame = {}
end

function CMD.initializing(pkg)
    luaver.init()
    ev.emit('initializing', pkg.config)
end

function CMD.initialized()
    initialized = true
    sendToMaster 'eventThread' {
        reason = 'started',
    }
end

function CMD.disconnect()
    if initialized then
        initialized = false
        state = 'running'
        ev.emit('terminated')
        sendToMaster 'eventThread' {
            reason = 'exited'
        }
    end
end

local function getFuncName(depth)
    local funcname = traceback.pushglobalfuncname(info.func)
    if funcname then
        return funcname
    end
    if info.what == 'main' then
        return '(main)'
    end
    if info.namewhat == '' then
        if luaver.LUAVERSION >= 52
            and info.what == "Lua"
            and rdebug.getinfo(depth, "t", info)
            and info.istailcall
        then
            return '(...tail calls...)'
        end
        local previous = {}
        if rdebug.getinfo(depth+1, "S", previous) then
            if previous.what == "Lua" or previous.what == "main" then
                return '(anonymous function)'
            end
            if previous.what == "C" then
                return '(called from C)'
            end
        end
        return ('(%s ?)'):format(info.what)
    end
    if info.namewhat == 'for iterator' then
        return '(for iterator)'
    end
    if info.namewhat == 'hook' then
        return '(hook)'
    end
    if info.namewhat == 'metamethod' then
        return ('(metamethod %s)'):format(info.name)
    end
    if info.namewhat == 'field' and info.name == 'integer index' then
        return '(field ?)'
    end
    if info.name == '?' then
        return ('(%s ?)'):format(info.namewhat)
    end
    return info.name
end

local function stackTrace(res, coid, start, levels)
    for depth = start, start + levels - 1 do
        if not rdebug.getinfo(depth, "Slnf", info) then
            return true, depth - start
        end
        local r = {
            id = (coid << 16) | depth,
            name = getFuncName(depth),
            line = 0,
            column = 0,
        }
        if info.what ~= 'C' then
            r.column = 1
            local src = source.create(info.source)
            if source.valid(src) then
                r.line = source.line(src, info.currentline)
                r.source = source.output(src)
                r.presentationHint = 'normal'
            else
                r.line = info.currentline
                r.presentationHint = 'label'
            end
        else
            r.presentationHint = 'label'
        end
        res[#res + 1] = r
    end
    return false, levels
end

local function skipCFunction(res)
    for i = #res, 1, -1 do
        if res[i].presentationHint ~= "label" or res[i].name ~= "(C ?)" then
            break
        end
        res[i] = nil
    end
end

function CMD.stackTrace(pkg)
    local start = pkg.startFrame and pkg.startFrame or 0
    local levels = (pkg.levels and pkg.levels ~= 0) and pkg.levels or 200
    local res = {}

    --
    -- 在VSCode的实现中这是一帧的第一个请求，所以在这里清理上一帧的数据。
    -- 很特殊，但目前也只能这样。
    --
    if start == 0 and levels == 1 then
        cleanFrame()
    end

    start = start + skipFrame
    local L = baseL
    local coroutineId = 0
    repeat
        hookmgr.sethost(L)
        local curL = L
        L = coroutineTree[curL]
        if stackFrame[curL] == nil then
            local finsh, n = stackTrace(res, coroutineId, start, levels)
            if not finsh then
                break
            end
            if not L then
                skipCFunction(res)
                break
            end
            stackFrame[curL] = start + n
            start = 0
            levels = levels - n
        elseif start > stackFrame[curL] then
            start = start - stackFrame[curL]
        end
        coroutineId = coroutineId + 1
    until (not L or levels <= 0)
    hookmgr.sethost(baseL)

    sendToMaster 'stackTrace' {
        command = pkg.command,
        seq = pkg.seq,
        success = true,
        body = {
            stackFrames = res,
            totalFrames = 0x10000,
        }
    }
end

function CMD.source(pkg)
    sendToMaster 'source' {
        command = pkg.command,
        seq = pkg.seq,
        content = source.getCode(pkg.sourceReference),
    }
end

local function findFrame(id)
    local L = baseL
    for _ = 1, id - 1 do
        if not L then
            return
        end
        L = coroutineTree[L]
    end
    return L
end

function CMD.scopes(pkg)
    local coid = (pkg.frameId >> 16) + 1
    local depth = pkg.frameId & 0xFFFF
    hookmgr.sethost(assert(findFrame(coid)))
    sendToMaster 'scopes' {
        command = pkg.command,
        seq = pkg.seq,
        body = {
            scopes = variables.scopes(depth)
        }
    }
end

function CMD.variables(pkg)
    local vars, err = variables.extand(pkg.valueId, pkg.filter, pkg.start, pkg.count)
    if not vars then
        sendToMaster 'variables' {
            command = pkg.command,
            seq = pkg.seq,
            success = false,
            message = err,
        }
        return
    end
    sendToMaster 'variables' {
        command = pkg.command,
        seq = pkg.seq,
        success = true,
        body = {
            variables = vars
        }
    }
end

function CMD.setVariable(pkg)
    local var, err = variables.set(pkg.valueId, pkg.name, pkg.value)
    if not var then
        sendToMaster 'setVariable' {
            command = pkg.command,
            seq = pkg.seq,
            success = false,
            message = err,
        }
        return
    end
    sendToMaster 'setVariable' {
        command = pkg.command,
        seq = pkg.seq,
        success = true,
        body = {
            value = var.value,
            type = var.type,
        }
    }
end

function CMD.evaluate(pkg)
    local depth = pkg.frameId & 0xFFFF
    local ok, result = evaluate.run(depth, pkg.expression, pkg.context)
    if not ok then
        sendToMaster 'evaluate' {
            command = pkg.command,
            seq = pkg.seq,
            success = false,
            message = result,
        }
        return
    end
    sendToMaster 'evaluate' {
        command = pkg.command,
        seq = pkg.seq,
        success = true,
        body = result
    }
end

function CMD.setBreakpoints(pkg)
    if noDebug or not source.valid(pkg.source) then
        return
    end
    breakpoint.set_bp(pkg.source, pkg.breakpoints, pkg.content)
end

function CMD.setFunctionBreakpoints(pkg)
    breakpoint.set_funcbp(pkg.breakpoints)
end

function CMD.setExceptionBreakpoints(pkg)
    breakpoint.setExceptionBreakpoints(pkg.arguments)
end

function CMD.exceptionInfo(pkg)
    sendToMaster 'exceptionInfo' {
        command = pkg.command,
        seq = pkg.seq,
        body = {
            breakMode = 'always',
            exceptionId = currentException.message,
            details = {
                stackTrace = currentException.trace,
            }
        }
    }
end

function CMD.loadedSources()
    source.all_loaded()
end

function CMD.stop(pkg)
    if noDebug then
        return
    end
    state = 'stopped'
    stopReason = pkg.reason -- entry or pause
    hookmgr.step_in()
end

function CMD.run()
    state = 'running'
    hookmgr.step_cancel()
end

function CMD.stepOver()
    if noDebug then
        return
    end
    state = 'stepOver'
    hookmgr.step_over()
end

function CMD.stepIn()
    if noDebug then
        return
    end
    state = 'stepIn'
    hookmgr.step_in()
end

function CMD.stepOut()
    if noDebug then
        return
    end
    state = 'stepOut'
    hookmgr.step_out()
end

function CMD.restartFrame()
    cleanFrame()
    sendToMaster 'eventStop' {
        reason = 'restart',
    }
end

function CMD.setSearchPath(pkg)
    local function set_search_path(name)
        if not pkg[name] or pkg[name] == json.null then
            return
        end
        local value = pkg[name]
        if type(value) == 'table' then
            local path = {}
            for _, v in ipairs(value) do
                if type(v) == "string" then
                    path[#path+1] = fs.nativepath(v)
                end
            end
            value = table.concat(path, ";")
        else
            value = fs.nativepath(value)
        end
        local visitor = rdebug.field(rdebug.field(rdebug._G, "package"), name)
        if not rdebug.assign(visitor, value) then
            return
        end
    end
    set_search_path "path"
    set_search_path "cpath"
end

function CMD.customRequestShowIntegerAsDec()
    variables.showIntegerAsDec()
    sendToMaster 'eventInvalidated' {
        areas = "variables",
    }
end

function CMD.customRequestShowIntegerAsHex()
    variables.showIntegerAsHex()
    sendToMaster 'eventInvalidated' {
        areas = "variables",
    }
end

local function runLoop(reason, level)
    baseL = hookmgr.gethost()
    --TODO: 只在lua栈帧时需要text？
    sendToMaster 'eventStop' (reason)
    skipFrame = level or 0

    while true do
        workerThreadUpdate(0.01)
        if state ~= 'stopped' then
            break
        end
    end
end

local event = {}

local function event_breakpoint(src, line)
    if not source.valid(src) then
        hookmgr.break_closeline()
        return
    end
    local bp = breakpoint.hit_bp(src, source.line(src, line))
    if bp then
        state = 'stopped'
        runLoop {
            reason = 'breakpoint',
            hitBreakpointIds = {bp.id}
        }
        return true
    end
end

function event.bp(line)
    if not initialized then return end
    rdebug.getinfo(0, "S", info)
    local src = source.create(info.source)
    event_breakpoint(src, line)
end

function event.funcbp(func)
    if not initialized then return end
    local bp = breakpoint.hit_funcbp(func)
    if bp then
        state = 'stopped'
        runLoop {
            reason = 'function breakpoint',
            hitBreakpointIds = {bp.id}
        }
    end
end

function event.step(line)
    if not initialized then return end
    rdebug.getinfo(0, "S", info)
    local src = source.create(info.source)
    if event_breakpoint(src, line) then
        return
    end
    if not source.valid(src) then
        return
    end
    workerThreadUpdate()
    if state == 'running' then
        return
    elseif state == 'stepOver' or state == 'stepOut' or state == 'stepIn' then
        state = 'stopped'
        stopReason = 'step'
        hookmgr.step_cancel()
    end
    if state == 'stopped' then
        runLoop {
            reason = stopReason
        }
    end
end

function event.newproto(proto, level)
    if not initialized then return end
    rdebug.getinfo(level, "S", info)
    local src = source.create(info.source)
    if not source.valid(src) then
        return false
    end
    return breakpoint.newproto(proto, src, info.linedefined.."-"..info.lastlinedefined)
end

function event.update()
    workerThreadUpdate()
end

function event.enable_update(flag)
    openUpdate = flag
    hookmgr.update_open(not noDebug and openUpdate)
end

function event.print(...)
    if not initialized then return end
    local res = {}
    local args = table.pack(...)
    for i = 1, args.n do
        res[#res + 1] = variables.tostring(args[i])
    end
    res = table.concat(res, '\t') .. '\n'
    rdebug.getinfo(1, "Sl", info)
    stdout(res, info)
    return true
end

function event.iowrite(...)
    if not initialized then return end
    local res = {}
    local args = table.pack(...)
    for i = 1, args.n do
        res[#res + 1] = variables.tostring(args[i])
    end
    res = table.concat(res, '\t')
    rdebug.getinfo(1, "Sl", info)
    stdout(res, info)
    return true
end

local ERREVENT_ERRRUN    <const> = 0x02
local ERREVENT_ERRSYNTAX <const> = 0x03
local ERREVENT_ERRMEM    <const> = 0x04
local ERREVENT_ERRERR    <const> = 0x05
local ERREVENT_PANIC     <const> = 0x10

local function GlobalFunction(name)
    return rdebug.value(rdebug.fieldv(rdebug._G, name))
end

local function getExceptionType(errcode)
    errcode = errcode & 0xF
    if errcode == ERREVENT_ERRRUN then
        if rdebug.getinfo(0, "Slf", info) then
            if info.what ~= 'C' then
                return "runtime"
            end
            local raisefunc = rdebug.value(info.func)
            if raisefunc == GlobalFunction "assert" then
                return "assert"
            end
            if raisefunc == GlobalFunction "error" then
                return "error"
            end
        end
        return "other"
    elseif errcode == ERREVENT_ERRSYNTAX then
        return "syntax"
    elseif errcode == ERREVENT_ERRMEM then
        return "unknown"
    elseif errcode == ERREVENT_ERRERR then
        return "unknown"
    else
        return "unknown"
    end
end

local function getExceptionCaught(errcode)
    if errcode & ERREVENT_PANIC ~= 0 then
        return "panic"
    end
    local pcall = GlobalFunction 'pcall'
    local xpcall = GlobalFunction 'xpcall'
    local level = 0
    while true do
        if not rdebug.getinfo(level, "f", info) then
            break
        end
        if level >= 100 then
            return 'native'
        end
        local f = rdebug.value(info.func)
        if f == pcall then
            return 'lua'
        end
        if f == xpcall then
            return 'lua'
        end
        level = level + 1
    end
    return 'native'
end

local function getExceptionFlags(errcode)
    return {
        getExceptionType(errcode),
        getExceptionCaught(errcode),
    }
end

local function runException(flags, errobj)
    local errmsg = variables.tostring(errobj)
    local level, message, trace = traceback.traceback(errmsg)
    if level < 0 then
        return
    end
    local bp = breakpoint.hitExceptionBreakpoint(flags, level, errobj)
    if not bp then
        return
    end
    currentException = {
        message = message,
        trace = trace,
    }
    state = 'stopped'
    runLoop({
        reason = 'exception',
        hitBreakpointIds = {bp.id},
        text = message,
    }, level)
end

function event.exception(errobj, errcode)
    if not initialized then return end
    if errcode == nil or errcode == -1 then
        --TODO:暂时兼容旧版本
        errcode = ERREVENT_ERRRUN
    end
    runException(getExceptionFlags(errcode), errobj)
end

function event.thread(co, type)
    local L = hookmgr.gethost()
    if co then
        if type == 0 then
            coroutineTree[L] = co
        elseif type == 1 then
            coroutineTree[co] = nil
        end
    end
    hookmgr.updatehookmask(L)
end

function event.wait()
    while not initialized do
        workerThreadUpdate(0.01)
    end
end

function event.setThreadName(name)
    sendToMaster 'setThreadName' (name)
end

function event.exit()
    sendToMaster 'exitWorker' {}
end

hookmgr.init(function(name, ...)
    local ok, e = xpcall(function(...)
        if event[name] then
            return event[name](...)
        end
    end, debug.traceback, ...)
    if not ok then
        err:push(e)
        return
    end
    return e
end)

local function lst2map(t)
    local r = {}
    for _, v in ipairs(t) do
        r[v] = true
    end
    return r
end

ev.on('initializing', function(config)
    noDebug = config.noDebug
    hookmgr.update_open(not noDebug and openUpdate)
    if hookmgr.thread_open then
        hookmgr.thread_open(true)
    end
    outputCapture = lst2map(config.outputCapture)
    if outputCapture["print"] then
        stdio.open_print(true)
    end
    if outputCapture["io.write"] then
        stdio.open_iowrite(true)
    end
end)

ev.on('terminated', function()
    hookmgr.step_cancel()
    if outputCapture["print"] then
        stdio.open_print(false)
    end
    if outputCapture["io.write"] then
        stdio.open_iowrite(false)
    end
end)

sendToMaster 'initWorker' {}

hookmgr.update_open(true)
